#include <stdio.h>
#include <stdarg.h>

extern int __dtostr(double d,char *buf,unsigned int maxlen,unsigned int prec,unsigned int prec2,int g);

typedef enum {
	ft_invalid,
	ft_char,
	ft_int,
	ft_long,
	ft_float,
	ft_str
} ftype_t;

static int __doprnt_strlen(const char *s)
{
	int l;
	for(l = 0; *s; s++) l++;
	return l;
}

#define isnum(c) (((c) >= '0') && ((c) <= '9'))

static int interpret_format(const char **format, char *fotype, ftype_t *fitype, int *unsign, int *left, int *w1, int *w2, int *zeropad)
{
	/* set up defaults */
	(*format)++;
	*unsign = 1;
	*left = 0;
	*w1 = -1;
	*w2 = -1;
	*zeropad = 0;
	*fitype = ft_invalid;
	*fotype = '\0';

	/* look for '-'left */
	if (**format == '-') {
		(*format)++;
		*left = 1;
	}

	/* look for zero padding */
	if (**format == '0') {
		(*format)++;
		*zeropad = 1;
	}

	/* look for a number */
	if (isnum(**format)) {
		*w1 = 0;
		while(isnum(**format)) {
			*w1 = *w1 * 10 + (**format - '0');
			(*format)++;
		}
	}

	/* look for a dot for decimal digits */
	if (**format == '.') {
		(*format)++;
		if (isnum(**format)) {
			*w2 = 0;
			while(isnum(**format)) {
				*w2 = *w2 * 10 + (**format - '0');
				(*format)++;
			}
		}
	}



	/* look for 'l', the long modifier */
	if (**format == 'l') {
		(*format)++;
		*zeropad = 1;
		*fitype = ft_long;
		*fotype = 'd';
	}

	switch(**format) {
		case 'd': /* decimal */
			*unsign = 0;
		case 'u': /* unsigned */
		case 'o': /* octal */
		case 'x': /* hexadecimal */
			if (*fitype == ft_invalid)
				*fitype = ft_int;
			*fotype = **format;
			(*format)++;
			break;
		case 'f': /* float */
		case 'e': /* float, exp */
		case 'g': /* float, autoselect */
			*fitype = ft_float;
			*fotype = **format;
			(*format)++;
			break;
		case 'c': /* character */
			*fitype = ft_char;
			*fotype = **format;
			(*format)++;
			break;
		case 's': /* string */
			*fitype = ft_str;
			*fotype = **format;
			(*format)++;
			break;
	}
	return (*fitype) == ft_invalid;
}

int long2str(char *buff, unsigned long i, int funsigned, int base)
{
	char *digits="0123456789abcdef";
	int d, slen, neg;

	slen = 0;
	neg = 0;
	if (!funsigned) {
		long int l;
		l = i;
		if (l < 0) {
			neg = 1;
			l = -l;
		}
		i = l;
	}

	if (i == 0) {
		*buff = '0';
		buff++;
		slen++;
	}
	else while(i > 0) {
		d = i % base;
		*buff = digits[d];
		buff++;
		i = i / base;
		slen++;
	}

	if (neg) {
		*buff = '-';
		buff++;
		slen++;
	}

	*buff = 0;
	return slen-1;
}


int __do_vsnprintf(char *out, int len, const char *format, va_list list)
{
	char *o;
	int l;

	char fotype;
	ftype_t fitype;
	int funsigned;
	int fleft;
	int fw1, fw2;
	int fzeropad;

	long int aint;
	double adouble;
	char achar;
	char *astr;

	char buff[128];
	int bl;

#define cappend(c) \
	do { \
		l++; \
		if (l >= len) \
			goto too_long; \
		*o = c; \
		o++; \
	} while(0)

#define srappend(str, slen, w, zeropad, fleft, psign) \
	do { \
		char *s; \
		int W; \
		if (!fleft) {\
			if ((psign) && (str[slen] == '-')) { \
				cappend(str[slen]); \
				slen--; \
				w--;\
			} \
			for(W = w-slen; W > 1; W--) \
				cappend(zeropad ? '0' : ' '); \
		} \
		for(s = str+slen; s >= str; s--) \
			cappend(*s); \
		if (fleft) \
			for(W = w-slen; W > 1; W--) \
				cappend(' '); \
	} while(0)

#define sappend(str, w, zeropad, fleft) \
	do { \
		char *s; \
		int W;\
		int slen = __doprnt_strlen(str); \
		if (!fleft) {\
			for(W = w-slen; W > 0; W--) \
				cappend(zeropad ? '0' : ' '); \
		} \
		for(s = str; *s; s++) \
			cappend(*s); \
		if (fleft) \
			for(W = w-slen; W > 0; W--) \
				cappend(' '); \
	} while(0)


	for(o = out, l = 0; *format != '\0';) {
		if (*format == '%') {
			if (interpret_format(&format, &fotype, &fitype, &funsigned, &fleft, &fw1, &fw2, &fzeropad) != 0)
				return -1;
			switch(fitype) {
				case ft_invalid:  return -1;
				case ft_char:     achar   = va_arg(list, int); break;
				case ft_int:      aint    = va_arg(list, int); break;
				case ft_long:     aint    = va_arg(list, long int); break;
				case ft_float:    adouble = va_arg(list, double); break;
				case ft_str:      astr    = va_arg(list, char *); break;
			}

			switch(fotype) {
				case 'u': /* unsigned */
				case 'd': /* decimal */
					bl = long2str(buff, aint, funsigned, 10);
					srappend(buff, bl, fw1, fzeropad, fleft, 1);
					break;
				case 'o': /* octal */
					bl = long2str(buff, aint, funsigned, 8);
					srappend(buff, bl, fw1, fzeropad, fleft, 1);
					break;
				case 'x': /* hexadecimal */
					bl = long2str(buff, aint, funsigned, 16);
					srappend(buff, bl, fw1, fzeropad, fleft, 1);
					break;
				case 'f': /* float */
					if (fw2 < 0)
						fw2 = 6;
					if (fw1 < 0)
						fw1 = 0;
					__dtostr(adouble, buff, sizeof(buff)-1, fw1, fw2, 0);
					bl = __doprnt_strlen(buff)-1;
					sappend(buff, fw1, fzeropad, fleft);
					break;
				case 'g': /* float, autoselect */
				case 'e': /* float, exp */
					if (fw2 < 0)
						fw2 = 6;
					if (fw1 < 0)
						fw1 = 0;
					{
						int e = 0;
						/* how ugly this can be... */
						while((adouble > 10) || (adouble < -10)) {
							adouble = adouble / 10;
							e++;
						}
						while((adouble < 1) && (adouble > -1)) {
							adouble = adouble * 10;
							e--;
						}
						__dtostr(adouble, buff, sizeof(buff), fw1, fw2, 0);
						bl = __doprnt_strlen(buff)-1;
						buff[bl++] = 'e';
						if (e < 0) {
							buff[bl++] = '-';
							e = -e;
						}
						else
							buff[bl++] = '+';
						buff[bl++] = (e / 10) + '0';
						buff[bl++] = (e % 10) + '0';
						buff[bl++] = '\0';
						fw1++;
						sappend(buff, fw1, fzeropad, fleft);

					}
					break;
				case 'c': /* character */
					buff[0] = achar;
					buff[1] = '\0';
					sappend(buff, fw1, 0, fleft);
					break;
				case 's': /* string */
					if (astr == NULL)
						astr = "(null)";
					sappend(astr, fw1, 0, fleft);
					break;
			}
		}
		else {
			cappend(*format);
			format++;
		}
	}
	cappend('\0');
	l--;
//va_arg(list, void *);

too_long:;
	return l;

#undef cappend
#undef srappend
#undef sappend
}

#undef isnum

int _doprnt(const char *format, va_list list, FILE *f)
{
	int len;
	char buf[32768], *s;

	for(s = buf, len = __do_vsnprintf(buf, sizeof(buf)-1, format, list); len > 0; len--,s++) {
		putc(*s, f);
	}
	return len;
}
