#include <stdarg.h>
#include <stdio.h>
//#include <stdlib.h>

extern int __do_vsnprintf(char *out, int len, const char *format, ...);

int __do_snprintf(char *out, int len, const char *format, ...)
{
	va_list l;
	int ret;

	va_start(l, format);
		ret = __do_vsnprintf(out, len, format, l);
	va_end(l);
	return ret;
}

#define tprintf(fmt, x...) \
	do { \
		char s[65536]; \
		printf(" %s ", fmt); \
		printf("%d ", __do_snprintf(s, sizeof(s)-1, fmt, x)); \
		printf("'%s'", s); \
		printf(" ref: "); \
		printf("%d ", snprintf(s, sizeof(s)-1, fmt, x)); \
		printf("'%s'\n", s); \
	} while(0)

int main()
{
	printf("No format:\n");
	tprintf("foo\tbar", 0);

	printf("Normal integer conversion:\n");
	tprintf("%d", 1);
	tprintf("%d", 123);
	tprintf("%o", 42);
	tprintf("%x", 42);

	printf("Signed/unsigned integer conversion:\n");
	tprintf("%d", -1);
	tprintf("%u", -1);
	tprintf("%ud", -1);
	tprintf("%d", 4295000000);
	tprintf("%u", 4295000000);

	printf("Padding integer conversion:\n");
	tprintf("%4d", 123);
	tprintf("%-4d", -1);
	tprintf("%04d", -1);
	tprintf("%-04d", -1);


	printf("Padding char conversion:\n");
	tprintf("@%c@", 'v');
	tprintf("@%4c@", 'w');
	tprintf("@%04c@", 'x');
	tprintf("@%-4c@", 'y');
	tprintf("@%-04c@", 'z');

	printf("Padding string conversion:\n");
	tprintf("@%s@", "hij");
	tprintf("@%6s@", "klm");
	tprintf("@%06s@", "nop");
	tprintf("@%-6s@", "qrs");
	tprintf("@%-06s@", "tuv");

	printf("Normal float conversion:\n");
	tprintf("%f", (float)3.14);
	tprintf("%12f", 3.14);
	tprintf("%4f", 3.14);
	tprintf("%4.2f", 3.14);
	tprintf("%4.5f", 3.14);
	tprintf("%-9f", 3.14);
	tprintf("%011f", 3.14);
	tprintf("%-011f", 3.14);

	printf("Exp float conversion:\n");
	tprintf("%e", (float).314);
	tprintf("%12e", 31.4);
	tprintf("%4e", 314);
	tprintf("%4.2e", .00314);
	tprintf("%4.5e", 3.14);
	tprintf("%-9e", 3.14);
	tprintf("%011e", 3.14);
	tprintf("%-011e", 3.14);

	return 0;
}
